""" AppOptics APM instrumentation for requests.

Instrumentation is done in urllib3.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
from appoptics_apm import util

HTTPLIB_LAYER = 'httplib'

appoptics_apm_logger = util.logger


def safeindex(_list, index):
    return _list[index] if len(_list) > index else None


def safeget(obj, key):
    return obj.get(key, None) if obj and hasattr(obj, 'get') else None


def wrap_request_putrequest(func, f_args, f_kwargs):
    self = safeindex(f_args, 0)
    if self:
        self.__appoptics_apm_path = safeindex(f_args, 2) or safeget(f_kwargs, 'url')
    return f_args, f_kwargs, {}


def wrap_request_endheaders(func, f_args, f_kwargs):
    if len(f_args) >= 1:
        self = f_args[0]
        self.putheader('X-Trace', util.last_id())
    return f_args, f_kwargs, {}


def wrap_request_getresponse(func, f_args, f_kwargs, res):
    kvs = { 'IsService': 'yes',
            'RemoteProtocol': 'http' }
    self = safeindex(f_args, 0)
    if self:
        kvs['RemoteHost'] = "%s:%s" % (getattr(self, 'host', ''), getattr(self, 'port', '80'))
        if hasattr(self, '__appoptics_apm_path'):
            kvs['ServiceArg'] = getattr(self, '__appoptics_apm_path') or '/'
            delattr(self, '__appoptics_apm_path')
    if hasattr(res, 'status'):
        kvs['HTTPStatus'] = getattr(res, 'status')
    edge_str = res.getheader('x-trace') if res and hasattr(res, 'getheader') else None
    return kvs, edge_str


def wrap(module):
    try:
        # Wrap putrequest.  This marks the beginning of the request, and is also
        # where
        wrapper_putrequest = util.log_method(
            HTTPLIB_LAYER, before_callback=wrap_request_putrequest,
            send_exit_event=False, store_backtrace=util._collect_backtraces('httplib')
        )
        setattr(module.HTTPConnection, 'putrequest',
                wrapper_putrequest(module.HTTPConnection.putrequest))

        wrapper_endheaders = util.log_method(
            HTTPLIB_LAYER, before_callback=wrap_request_endheaders,
            send_entry_event=False, send_exit_event=False
        )
        setattr(module.HTTPConnection, 'endheaders',
                wrapper_endheaders(module.HTTPConnection.endheaders))

        wrapper_getresponse = util.log_method(
            HTTPLIB_LAYER, callback=wrap_request_getresponse, send_entry_event=False
        )
        setattr(module.HTTPConnection, 'getresponse',
                wrapper_getresponse(module.HTTPConnection.getresponse))
    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))


if util.ready():
    try:
        import httplib
        wrap(httplib)
    except ImportError:
        import http.client
        wrap(http.client)



"""AppOptics APM instrumentation for Django

 Copyright (C) 2016 by SolarWinds, LLC.
 All rights reserved.
"""
import time
import functools
import threading
from appoptics_apm import util
from appoptics_apm import imports
try:
    from django.utils.deprecation import MiddlewareMixin as AppOpticsApmMiddlewareBase
except ImportError:
    AppOpticsApmMiddlewareBase = object

# django middleware for passing values to appoptics_apm
__all__ = ("AppOpticsApmDjangoMiddleware", "install_appoptics_apm_middleware")

appoptics_apm_logger = util.logger


class AppOpticsApmWSGIHandler(object):
    """ Wrapper WSGI Handler for Django's django.core.handlers.wsgi:WSGIHandler
    Can be used as a replacement for Django's WSGIHandler, e.g. with uWSGI.
    """
    def __init__(self):
        """ Import and instantiate django.core.handlers.WSGIHandler,
        now that the load_middleware wrapper below has been initialized. """
        from django.core.handlers.wsgi import WSGIHandler
        self._handler = WSGIHandler()

    def __call__(self, environ, start_response):
        return self._handler(environ, start_response)


# Middleware hooks listed here: http://docs.djangoproject.com/en/dev/ref/middleware/
class AppOpticsApmDjangoMiddleware(AppOpticsApmMiddlewareBase):
    def __init__(self, *args, **kwargs):
        from django.conf import settings
        try:
            self.layer = settings.APPOPTICS_APM_BASE_LAYER
        except AttributeError:
            self.layer = 'django'
        super(AppOpticsApmDjangoMiddleware, self).__init__(*args, **kwargs)

    def _singleline(self, e):  # some logs like single-line errors better
        return str(e).replace('\n', ' ').replace('\r', ' ')

    def process_request(self, request):
        try:
            xtr_hdr = request.META.get("HTTP_X-Trace",   request.META.get("HTTP_X_TRACE"))
            ctx = util.start_trace(self.layer, xtr=xtr_hdr, store_backtrace=False)
            request.META.setdefault('APPOPTICS_APM_SPAN_START', str(int(time.time() * 1000000)))

        except Exception as e:
            appoptics_apm_logger.error("AppOptics APM middleware error: %s" % self._singleline(e))

    def process_view(self, request, view_func, view_args, view_kwargs):
        if not util.Context.get_default().is_sampled():
            return
        try:
            controller = getattr(view_func, '__module__', None)
            action = getattr(view_func, '__name__', None)

            kvs = {
                'Controller': controller,
                'Action': action
            }
            util.log('process_view', None, keys=kvs, store_backtrace=False)
            if controller and action:
                request.META.setdefault('APPOPTICS_APM_SPAN_TRANSACTION', '{c}.{a}'.format(c=controller, a=action))
        except Exception as e:
            appoptics_apm_logger.error("AppOptics APM middleware error: %s" % self._singleline(e))

    def process_response(self, request, response):
        """Process the response, record some information and send the end_trace out
        """
        ctx = util.Context.get_default()

        if ctx  and not ctx.is_sampled():
            response['X-Trace'] = str(ctx)
            return response

        try:
            kvs = None
            transaction_name = ctx.get_transaction_name()
            if not transaction_name:
                transaction_name = request.META.pop('APPOPTICS_APM_SPAN_TRANSACTION', None)
            # exit event needs it. will be processed there
            domain =request.META.get('HTTP_X_FORWARDED_HOST', request.META.get('HTTP_HOST', 'localhost'))

            setattr(util.Context.transaction_dict, 'url_tran', request.build_absolute_uri())
            setattr(util.Context.transaction_dict, 'domain', domain)
            setattr(util.Context.transaction_dict, 'transaction_name', transaction_name)
            setattr(util.Context.transaction_dict, 'start_time', request.META.pop('APPOPTICS_APM_SPAN_START', None))
            setattr(util.Context.transaction_dict, 'request_method', request.META.get('REQUEST_METHOD', 'GET'))
            setattr(util.Context.transaction_dict, 'status_code', response.status_code)
            x_trace = util.end_http_trace(self.layer, keys=kvs)

            if x_trace:
                response['X-Trace'] = x_trace
                appoptics_apm_logger.debug("djangoware process_response x_trace: {x}".format( x = x_trace))
        except Exception as e:
            appoptics_apm_logger.error("AppOptics APM middleware error: %s" % self._singleline(e))

        return response

    def process_exception(self, request, exception):
        try:
            util.log_exception()
        except Exception as e:
            appoptics_apm_logger.error("AppOptics APM middleware error: %s" % self._singleline(e))


def middleware_hooks(module, objname):
    try:
        # wrap middleware callables we want to wrap
        cls = getattr(module, objname, None)
        appoptics_apm_logger.info('middleware_hooks module {m}'.format(m=module))
        if not cls:
            return
        for method in ['process_request',
                       'process_view',
                       'process_response',
                       'process_template_response',
                       'process_exception']:
            fn = getattr(cls, method, None)
            if not fn:
                continue
            profile_name = '%s.%s.%s' % (module.__name__, objname, method)
            setattr(cls, method,
                    util.profile_function(profile_name)(fn))
    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))

load_middleware_lock = threading.Lock()


def on_load_middleware():
    """ wrap Django middleware from a list """

    # protect middleware wrapping: only a single thread proceeds
    global load_middleware_lock         # lock gets overwritten as None after init
    if not load_middleware_lock:        # already initialized? abort
        return
    mwlock = load_middleware_lock
    mwlock.acquire()                    # acquire global lock
    if not load_middleware_lock:        # check again
        mwlock.release()                # abort
        return
    load_middleware_lock = None         # mark global as "init done"

    try:
        # middleware hooks
        from django.conf import settings

        # settings.MIDDLEWARE_CLASSES is a tuple versions prior 1.9 and is list
        # in later versions. Versions since 1.10 use MIDDLEWARE instead of MIDDLEWARE_CLASSES.
        # however both attributes may exist in versions 1.10 and above.
        # The real type and value of settings.MIDDLEWARE and
        # settings.MIDDLEWARE_CLASSES are determined by the value in setting
        # file. Django.core.handler(1.11.16): load_middleware checks MIDDLEWARE first,
        # If it is none, it uses MIDDLEWARE_CLASSES, otherwise it uses
        # MIDDLEWARE. It doesn't check it's tuple or list, just iterate it.
        using_middleware_attr = True
        middleware = getattr(settings, 'MIDDLEWARE', None)
        if middleware is None:
            middleware = getattr(settings, 'MIDDLEWARE_CLASSES', None)
            using_middleware_attr = False


        for i in [] if middleware is None else middleware:
            if i.startswith('appoptics_apm'):
                continue
            dot = i.rfind('.')
            if dot < 0 or dot+1 == len(i):
                continue
            objname = i[dot+1:]
            imports.whenImported(i[:dot],
                                 functools.partial(middleware_hooks, objname=objname))
        # ORM
        if util.config['inst_enabled']['django_orm']:
            from appoptics_apm import inst_django_orm
            # The wrapper path BaseDatabaseWrapper has changed in Django 1.8 and onwards.
            try:
                import django.db.backends.base.base
                imports.whenImported('django.db.backends.base.base', inst_django_orm.wrap)
            except ImportError as e:
                appoptics_apm_logger.debug('AppOptics on_load_middleware: {e}, try loading dummy.base'.format(e=e))
                try:
                    import django.db.backends.dummy.base
                    imports.whenImported('django.db.backends.dummy.base', inst_django_orm.wrap)
                except ImportError as e:
                    appoptics_apm_logger.error('AppOptics error in on_load_middleware: {e}'.format(e=e))

        # templates
        if util.config['inst_enabled']['django_templates']:
            from appoptics_apm import inst_django_templates
            import django
            imports.whenImported('django.template.base', inst_django_templates.wrap)

        # load pluggable instrumentation
        from .loader import load_inst_modules
        load_inst_modules()
        apm_middleware = 'appoptics_apm.djangoware.AppOpticsApmDjangoMiddleware'
        if isinstance(middleware, list):
            middleware.insert(0, apm_middleware)
        elif isinstance(middleware, tuple):
            if using_middleware_attr:
                settings.MIDDLEWARE = (apm_middleware,) + middleware
            else:
                settings.MIDDLEWARE_CLASSES = (apm_middleware,) + middleware
        else:
            appoptics_apm_logger.error(
                "AppOptics APM error: settings middleware attribute should be either a tuple or a list,"
                "got {mw_type}".format(mw_type=str(type(middleware)))
            )
    except Exception as e:
        appoptics_apm_logger.error('AppOptics APM error in on_load_middleware: {e}'.format(e=e))

    finally:  # release instrumentation lock
        mwlock.release()


def install_appoptics_apm_middleware(module):
    def base_handler_wrapper(func):
        @functools.wraps(func)  # XXX Not Python2.4-friendly
        def wrap_method(*f_args, **f_kwargs):
            on_load_middleware()
            return func(*f_args, **f_kwargs)
        return wrap_method

    try:
        cls = getattr(module, 'BaseHandler', None)
        try:
            if not cls or cls.APPOPTICS_APM_MIDDLEWARE_LOADER:
                return
        except AttributeError as e:
            cls.APPOPTICS_APM_MIDDLEWARE_LOADER = True
        fn = getattr(cls, 'load_middleware', None)
        setattr(cls, 'load_middleware', base_handler_wrapper(fn))
    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: %s" % str(e))

if util.ready():
    try:
        imports.whenImported('django.core.handlers.base', install_appoptics_apm_middleware)
        # phone home
        util.report_layer_init(layer='django')
    except ImportError as e:
        # gracefully disable tracing if AppOptics APM lib not present
        appoptics_apm_logger.error("[AppOptics] Unable to instrument app and middleware: %s" % e)



""" WSGI middleware for appoptics_apm support

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""

from appoptics_apm import util
import sys
import time
from six.moves import urllib
from appoptics_apm.loader import load_inst_modules
import traceback as tb

MODULE_INIT_REPORTED = False


class AppOpticsApmMiddleware(object):
    def __init__(self, app, appoptics_apm_config=None, layer="wsgi", profile=False):
        """
        Install instrumentation for tracing a WSGI app.

        Arguments:
            app - the WSGI app that we're wrapping
            appoptics_apm_config - (optional) dictionary with appoptics_apm configuration parameters:
              - appoptics_apm.tracing_mode: 'always', 'never'
              - appoptics_apm.sample_rate: a number from 0 to 1000000 denoting fraction of requests to trace
            layer - (optional) layer name to use, default is "wsgi"
            profile - (optional) profile entire calls to app (don't use in production)
        """
        if appoptics_apm_config is None:
            appoptics_apm_config = {}

        self.wrapped_app = app
        self.appoptics_apm_config = appoptics_apm_config
        self.layer = layer
        self.profile = profile

        if self.appoptics_apm_config.get('appoptics_apm.tracing_mode'):
            util.config['tracing_mode'] = self.appoptics_apm_config['appoptics_apm.tracing_mode']

        if self.appoptics_apm_config.get('appoptics_apm.reporter_port'):
            util.config['reporter_port'] = self.appoptics_apm_config['appoptics_apm.reporter_port']

        if self.appoptics_apm_config.get('appoptics_apm.sample_rate'):
            util.config['sample_rate'] = float(self.appoptics_apm_config['appoptics_apm.sample_rate'])

        # load pluggable instrumentation
        load_inst_modules()

        # phone home
        global MODULE_INIT_REPORTED
        if not MODULE_INIT_REPORTED:
            util.report_layer_init(layer=layer)
            MODULE_INIT_REPORTED = True

    def __call__(self, environ, start_response):
        xtr_hdr = environ.get("HTTP_X-Trace", environ.get("HTTP_X_TRACE"))
        end_evt = None
        if not util.ready():
            return self.wrapped_app(environ, start_response)

        start_time = time.time() * 1e6
        # start the trace: ctx.is_valid() will be False if not tracing this request
        ctx, start_evt = util.Context.start_trace(self.layer, xtr=xtr_hdr)

        if ctx.is_sampled():
            # get some HTTP details from WSGI vars
            # http://www.wsgi.org/en/latest/definitions.html
            for hosthdr in ("HTTP_HOST", "HTTP_X_HOST", "HTTP_X_FORWARDED_HOST", "SERVER_NAME"):
                if hosthdr in environ:
                    start_evt.add_info("HTTP-Host", environ[hosthdr])
                    break
            if 'PATH_INFO' in environ:
                start_evt.add_info("URL", environ['PATH_INFO'])
            if 'REQUEST_METHOD' in environ:
                start_evt.add_info("Method", environ['REQUEST_METHOD'])
            if 'QUERY_STRING' in environ:
                start_evt.add_info("Query-String", environ['QUERY_STRING'])

            ctx.report(start_evt)
            ctx.set_as_default()
            end_evt = ctx.create_event('exit', self.layer)

        response_body = []

        # Seems I have to use a list for status_str as there is no 'non-local' in Python 2.
        status_str = ['200 OK', ]

        def wrapped_start_response(status, headers, exc_info=None):
            status_str[0] = status

            if ctx.is_sampled():
                end_evt.add_info("Status", status.split(' ', 1)[0])
                if exc_info:
                    _t, exc, trace = exc_info
                    end_evt.add_info("ErrorMsg", str(exc))
                    end_evt.add_info("ErrorClass", exc.__class__.__name__)
                    end_evt.add_info("Backtrace", "".join(tb.format_list(tb.extract_tb(trace))))
                headers.append(("X-Trace", end_evt.id()))
            elif ctx.is_valid():
                headers.append(("X-Trace", str(ctx)))
            start_response(status, headers)
            if self.profile:
                return response_body.append

        stats = None
        result = None
        exc_thrown = False
        domain = None

        try:
            if self.profile and ctx.is_sampled():
                try:
                    # XXX test cProfile and pstats exist
                    import cStringIO, cProfile, pstats
                except ImportError:
                    self.profile = False

            if self.profile and ctx.is_sampled():
                def runapp():
                    appiter = self.wrapped_app(environ, wrapped_start_response)
                    response_body.extend(appiter)
                    if hasattr(appiter, 'close'):
                        appiter.close()

                p = cProfile.Profile()
                p.runcall(runapp)
                body = ''.join(response_body)
                result = [body]

                try:
                    sio = cStringIO.StringIO()
                    s = pstats.Stats(p, stream=sio)
                    s.sort_stats('cumulative')
                    s.print_stats(15)
                    stats = sio.getvalue()
                    sio.close()
                except Exception as aoe:  # catch possible non-application thrown exception
                    util.logger.error('AppOpticsMiddleware cProfile, pstats'
                            'processing encountered problem:: {}'.format(aoe))
            else:
                result = self.wrapped_app(environ, wrapped_start_response)

        except Exception as e:
            exc_thrown = True
            util.logger.error('Error in AppOpticsApmMiddleware: {}'.format(e))
            raise    # re-throw wrapped app raised exception

        finally:
            trans_name = util.get_transaction_name()
            if end_evt:
                # check current TLS context and add to end event if valid
                if util.Context.get_default().is_sampled():
                    end_evt.add_edge(util.Context.get_default())

            #domain and url_tran are required for creating spans
            stats = None if exc_thrown else stats
            setattr(util.Context.transaction_dict, 'status_code', int(status_str[0].split()[0]))
            if util.config['transaction']['prepend_domain_name']:
                domain =environ.get('HTTP_X_FORWARDED_HOST', environ.get('HTTP_HOST'))
            setattr(util.Context.transaction_dict, 'url_tran', None if trans_name else self.get_url(environ))

            setattr(util.Context.transaction_dict, 'http', True)
            setattr(util.Context.transaction_dict, 'domain', domain)
            setattr(util.Context.transaction_dict, 'transaction_name', trans_name)
            setattr(util.Context.transaction_dict, 'start_time', start_time )
            setattr(util.Context.transaction_dict, 'request_method', environ.get('REQUEST_METHOD'))
            self.send_end(ctx, end_evt, environ, threw_error=exc_thrown, stats=stats)

        return result

    def get_url(self, environ):
        """
        Construct the url from environ, see PEP-0333.
        :param environ:
        :return: url
        """
        scheme = environ['wsgi.url_scheme']

        if environ.get('HTTP_HOST'):
            host = environ['HTTP_HOST']
        else:
            host = environ['SERVER_NAME']

        port = ''
        if environ['wsgi.url_scheme'] == 'https':
            if environ['SERVER_PORT'] != '443':
                port = ':{p}'.format(p=environ['SERVER_PORT'])
        else:
            if environ['SERVER_PORT'] != '80':
                port = ':{p}'.format(p=environ['SERVER_PORT'])

        if environ.get('QUERY_STRING'):
            query_string = '?{q}'.format(q=environ['QUERY_STRING'])
        else:
            query_string = ''

        if port in host:
            host_port = host
        else:
            host_port = '{host}{port}'.format(host=host, port=port)

        return '{scheme}://{host_port}{script_name}{path_info}{query_string}'.format(
            scheme=scheme, host_port=host_port,
            script_name=urllib.parse.quote(environ.get('SCRIPT_NAME', '')),
            path_info=urllib.parse.quote(environ.get('PATH_INFO', '')),
            query_string=query_string
        )

    @classmethod
    def send_end(cls, ctx, evt, environ, threw_error=None, stats=None):
        if ctx is None:
            util.logger.warning('ctx is None')
            return

        try:
            if ctx.is_sampled():
                evt.add_edge(ctx)
                util.logger.debug('AppOpticsMiddleware send event: {e}'.format(e = evt))
                if stats:
                    evt.add_info("Profile", stats)
                if threw_error:
                    _t, exc, trace = sys.exc_info()
                    evt.add_info("ErrorMsg", str(exc))
                    evt.add_info("ErrorClass", exc.__class__.__name__)
                    evt.add_info("Backtrace", "".join(tb.format_list(tb.extract_tb(trace))))
                    del trace  # delete reference to traceback object to allow garbage collection

                # gets controller, action
                for k, v in environ.get('wsgiorg.routing_args', [{}, {}])[1].items():
                    if k == "action":
                        evt.add_info(str(k).capitalize(), str(v))
                    elif k == "controller":
                        try:
                            # handle cases used in openstack's WSGI (and possibly others)
                            if v.controller:
                                evt.add_info(str(k).capitalize(), str(v.controller.__class__.__name__))
                            else:
                                evt.add_info(str(k).capitalize(), str(v))
                        except Exception:
                            evt.add_info(str(k).capitalize(), str(v))

        finally:
            # report, then clear trace context now that trace is over. Either ctx is_sampled or not,
            # end_trace must be called to create spans
            ctx.end_trace(evt)



""" AppOptics instrumentation API for Python.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.

appoptics_apm_noop defines no-op classes for platforms we don't support building the c extension on
"""

# No-op classes intentionally left undocumented
# "Missing docstring"
# pylint: disable-msg=C0103
import threading

class Metadata(object):
    def __init__(self, _=None):
        pass

    @staticmethod
    def fromString(_):
        return Metadata()

    def createEvent(self):
        return Event()

    @staticmethod
    def makeRandom(flag=True):
        return Metadata(True)

    def copy(self):
        return self

    def isValid(self):
        return False

    def isSampled(self):
        return False

    def toString(self):
        return ''

class Context(object):
    transaction_dict = threading.local()

    @staticmethod
    def init(_, __, ___):
        pass

    @staticmethod
    def setTracingMode(_):
        return False

    @staticmethod
    def setDefaultSampleRate(_):
        return False

    @staticmethod
    def sampleRequest(_, __):
        return False

    @staticmethod
    def get():
        return Metadata()

    @staticmethod
    def set(_):
        pass

    @staticmethod
    def fromString(_):
        return Context()

    @staticmethod
    def copy():
        return Context()

    @staticmethod
    def clear():
        pass

    @staticmethod
    def isValid():
        return False

    @staticmethod
    def toString():
        return ''

    @staticmethod
    def createEvent():
        return Event()

    @staticmethod
    def startTrace(_=None):
        return Event()

    @staticmethod
    def isReady(_):
        #unknown error
        return 0


class Event(object):
    def __init__(self, _=None, __=None):
        pass

    def addInfo(self, _, __):
        pass

    def addEdge(self, _):
        pass

    def addEdgeStr(self, _):
        pass

    def getMetadata(self):
        return Metadata()

    def metadataString(self):
        return ''

    def is_valid(self):
        return False

    @staticmethod
    def startTrace(_=None):
        return Event()


class Reporter(object):
    def __init__(self, _, __=None):
        pass

    def sendReport(self, _, __=None):
        pass

    def sendStatus(self, _, __=None):
        pass


class UdpReporter(object):
    def __init__(self, _, __=None):
        pass

    def sendReport(self, _, __=None):
        pass

    def sendStatus(self, _, __=None):
        pass

class SslReporter(object):
    def __init__(self, _, __=None):
        pass

    def sendReport(self, _, __=None):
        pass

    def sendStatus(self, _, __=None):
        pass

class Span(object):
    @staticmethod
    def createHttpSpan(_,__, ___, ____, _____, ______, _______):
        pass
    @staticmethod
    def createSpan(_, __, ___):
        pass

class MetricTags(object):
    def __init__(self, count):
        super(MetricTags, self).__init__()
        pass
    def add(*args, **kwargs):
        pass

class CustomMetrics(object):
    @staticmethod
    def summary(*args, **kwargs):
        pass

    @staticmethod
    def increment(*args, **kwargs):
        pass


class oboe_metric_tag_t(object):
    def __init__(self, k, v):
        self.key = k
        self.value = v


class DebugLog(object):
    @staticmethod
    def getLevelName(*args, **kwargs):
        pass

    @staticmethod
    def getModuleName(*args, **kwargs):
        pass

    @staticmethod
    def getLevel(*args, **kwargs):
        pass

    @staticmethod
    def setLevel(*args, **kwargs):
        pass

    @staticmethod
    def setOutputStream(*args, **kwargs):
        pass

    @staticmethod
    def setOutputFile(*args, **kwargs):
        pass

    @staticmethod
    def addDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def removeDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def logMessage(*args, **kwargs):
        pass



""" AppOptics APM instrumentation for redis client module.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
from functools import wraps
from appoptics_apm import util
import operator
import sys
from distutils.version import LooseVersion

appoptics_apm_logger = util.logger
# which commands can we get hit/miss info from?
# also supports mget and hmget separately, see below
HITTABLE_COMMANDS = set(('GET', 'GETSET', 'HGET', 'LINDEX', 'LGET',
                         'RPOPLPUSH', 'LPOP', 'RPOP', 'BRPOPLPUSH',
                         'SPOP', 'SRANDMEMBER'))

# which commands can we get KVKey from?
# for S*/H*/Z* commans, the key is currently the set/hash name
KEYABLE_COMMANDS = {'APPEND': 1,
                    'BITCOUNT': 1,
                    'BITOP': 2,  # XXX DEST
                    'DECR': 1,
                    'EXISTS': 1,
                    'EXPIRE': 1,
                    'EXPIREAT': 1,
                    'GET': 1,
                    'GETBIT': 1,
                    'GETRANGE': 1,
                    'GETSET': 1,
                    'INCR': 1,
                    'INCRBY': 1,
                    'INCRBYFLOAT': 1,
                    'MOVE': 1,
                    'PEXPIRE': 1,
                    'PEXPIREAT': 1,
                    'PSETEX': 1,
                    'PTTL': 1,
                    'RENAME': 1,
                    'RENAMEX': 1,
                    'SET': 1,
                    'SETBIT': 1,
                    'SETEX': 1,
                    'SETNX': 1,
                    'SETRANGE': 1,
                    'STRLEN': 1,
                    'SUBSTR': 1,
                    'TTL': 1,
                    'TYPE': 1,

                    'LINDEX': 1,
                    'LINSERT': 1,
                    'LLEN': 1,
                    'LPOP': 1,
                    'LPUSH': 1,
                    'LPUSHX': 1,
                    'LRANGE': 1,
                    'LREM': 1,
                    'LSET': 1,
                    'LTRIM': 1,
                    'RPOP': 1,
                    'RPUSH': 1,
                    'RPOPLPUSH': 1,
                    'RPUSHX': 1,
                    'SORT': 1,

                    'SADD': 1,
                    'SCARD': 1,
                    'SDIFF': 1,
                    'SDIFFSTORE': 1,  # XXX DEST
                    'SINTERSTORE': 1,  # XXX DEST
                    'SISMEMBER': 1,
                    'SMEMBERS': 1,
                    'SMOVE': 1,
                    'SPOP': 1,
                    'SRANDMEMBER': 1,
                    'SREM': 1,
                    'SUNIONSTORE': 1,  # XXX DEST

                    'ZADD': 1,
                    'ZCARD': 1,
                    'ZCOUNT': 1,
                    'ZINCRBY': 1,
                    'ZINTERSTORE': 1,  # XXX DEST
                    'ZRANGE': 1,
                    'ZRANGEBYSCORE': 1,
                    'ZRANK': 1,
                    'ZREM': 1,
                    'ZREMRANGEBYRANK': 1,
                    'ZREMRANGEBYSCORE': 1,
                    'ZREVRANGE': 1,
                    'ZREVRANGEBYSCORE': 1,
                    'ZREVRANK': 1,
                    'ZSCORE': 1,
                    'ZUNIONSCORE': 1,  # XXX DEST

                    'HDEL': 1,
                    'HEXISTS': 1,
                    'HGET': 1,
                    'HGETALL': 1,
                    'HINCRBY': 1,
                    'HINCRBYFLOAT': 1,
                    'HKEYS': 1,
                    'HLEN': 1,
                    'HSET': 1,
                    'HSETNX': 1,
                    'HMSET': 1,
                    'HMGET': 1,
                    'HVALS': 1,

                    'PUBLISH': 1  # CHANNEL
                    }

SCRIPT_COMMANDS = {'EVAL': 1,  # SCRIPT
                   'EVALSHA': 1  # SHA
                   }

# these commands have two parts, as separate args in python redis client
# eg. SCRIPT LOAD
TWO_PARTERS = set(('SCRIPT', 'CLIENT', 'CONFIG', 'DEBUG'))


def wrap_execute_command(func, f_args, f_kwargs, return_val):
    """ This is where most "normal" redis commands are instrumented. """
    kvs = {}

    # command
    op = f_args[1]
    if op in TWO_PARTERS:
        op = op + ' ' + f_args[2]
    kvs['KVOp'] = op.lower()

    # key
    if op in KEYABLE_COMMANDS:
        kvs['KVKey'] = f_args[1 + KEYABLE_COMMANDS[op]]

    # script eval
    if op in SCRIPT_COMMANDS:
        kvs['Script'] = f_args[1 + SCRIPT_COMMANDS[op]][0:100]  # max out at 100 chars if script

    # hit/miss
    if op in HITTABLE_COMMANDS:
        kvs['KVHit'] = return_val != None
    elif op == 'MGET' or op == 'HMGET':
        kvs['KVKeyCount'] = len(return_val)
        kvs['KVHitCount'] = sum(r != None for r in return_val)

    return kvs


def wrap_execute_pipeline(func, f_args, f_kwargs, return_val):
    # args: self, connection, commands, raise_on_error
    kvs = {}
    fp_cmds = {}
    load_redis = sys.modules['redis']
    if load_redis is None:
       logger.warning('Failed to load module: {m}'.format(m='redis'))
       return kvs 

    #redis version[2.4.0, 2.6.0] process args different from [2.6.0, ]
    if LooseVersion(load_redis.__version__) < LooseVersion("2.6.0") and \
       LooseVersion(load_redis.__version__) >= LooseVersion("2.4.0"):
        for (args, options) in f_args[1]:
            fp_cmds[args[0]] = fp_cmds.get(args[0], 0) + 1

        op = 'PIPE:' + ','.join([cmd for (cmd, _) in sorted(fp_cmds.items(), key=operator.itemgetter(0))])
        kvs['KVOp'] = op.lower()
    else:
        for (args, options) in f_args[2]:
            fp_cmds[args[0]] = fp_cmds.get(args[0], 0) + 1

        op = 'PIPE:' + ','.join([cmd for (cmd, _) in sorted(fp_cmds.items(), key=operator.itemgetter(0))])
        kvs['KVOp'] = op.lower()

    return kvs


def wrap_send_packed_command(layer_name, func):
    """ This is where we get the RemoteHost.

    This relies on the module internals, and just sends an info event when this
    function is called.
    """

    @wraps(func)  # XXX Not Python2.4-friendly
    def wrapper(*f_args, **f_kwargs):
        ret = func(*f_args, **f_kwargs)
        try:
            conn_obj = f_args[0]
            if 'path' in dir(conn_obj):
                host = 'localhost'
            else:
                host = conn_obj.host + ':' + str(conn_obj.port)

            util.log('info', layer_name, keys={'RemoteHost': host},
                     store_backtrace=util._collect_backtraces('redis'))
        except Exception as e:
            appoptics_apm_logger.error('AppOptics APM error: {e}'.format(e=e))
        return ret

    return wrapper


def wrap(layer_name, module):
    try:
        # first get the basic client methods; common point of processing is execute_command
        # client is StrictRedis for >= 2.4.10, or Redis for < 2.4.10
        cls = getattr(module, 'StrictRedis', getattr(module, 'Redis', None))
        if cls:
            execute_command = cls.execute_command
            wrapper = util.log_method(layer_name,
                                      callback=wrap_execute_command)
            setattr(cls, 'execute_command', wrapper(execute_command))
        else:
            appoptics_apm_logger.error("AppOptics APM error: couldn't find redis.StrictRedis nor redis.Redis class to"
                              " instrument, redis coverage may be partial.")

        # RemoteHost
        cls = getattr(module, 'Connection', None)
        if cls:
            wrapper = wrap_send_packed_command(layer_name, cls.send_packed_command)
            setattr(cls, 'send_packed_command', wrapper)
        else:
            appoptics_apm_logger.error("AppOptics APM error: couldn't find redis.Connection class to instrument, "
                              "redis coverage may be partial.")

        # pipeline/multi
        cls = getattr(module.client, 'BasePipeline', getattr(module.client, 'Pipeline', None))
        if cls:
            wrapper = util.log_method(layer_name, callback=wrap_execute_pipeline)
            execute_pipeline = cls._execute_pipeline
            setattr(cls, '_execute_pipeline', wrapper(execute_pipeline))
            execute_transaction = cls._execute_transaction
            setattr(cls, '_execute_transaction', wrapper(execute_transaction))
        else:
            appoptics_apm_logger.error("AppOptics APM error: couldn't find redis.client.BasePipeline nor "
                              "redis.client.Pipeline class to instrument, "
                              "redis coverage may be partial.")

        # pubsub
        cls = getattr(module.client, 'PubSub', None)
        if cls:
            execute_command = cls.execute_command
            wrapper = util.log_method(layer_name, callback=wrap_execute_command)
            setattr(cls, 'execute_command', wrapper(execute_command))
        else:
            appoptics_apm_logger.error("AppOptics APM error: couldn't find redis.client.PubSub class to instrument, "
                              "redis coverage may be partial.")

    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM error: {e}".format(e=e))


if util.ready():
    try:
        import redis

        wrap('redis', redis)
    except ImportError as ex:
        pass



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_oboe')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_oboe')
    _oboe = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_oboe', [dirname(__file__)])
        except ImportError:
            import _oboe
            return _oboe
        try:
            _mod = imp.load_module('_oboe', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _oboe = swig_import_helper()
    del swig_import_helper
else:
    import _oboe
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0

OBOE_DEBUG_DISABLED = _oboe.OBOE_DEBUG_DISABLED
OBOE_DEBUG_FATAL = _oboe.OBOE_DEBUG_FATAL
OBOE_DEBUG_ERROR = _oboe.OBOE_DEBUG_ERROR
OBOE_DEBUG_WARNING = _oboe.OBOE_DEBUG_WARNING
OBOE_DEBUG_INFO = _oboe.OBOE_DEBUG_INFO
OBOE_DEBUG_LOW = _oboe.OBOE_DEBUG_LOW
OBOE_DEBUG_MEDIUM = _oboe.OBOE_DEBUG_MEDIUM
OBOE_DEBUG_HIGH = _oboe.OBOE_DEBUG_HIGH
OBOE_MODULE_ALL = _oboe.OBOE_MODULE_ALL
OBOE_MODULE_UNDEF = _oboe.OBOE_MODULE_UNDEF
OBOE_MODULE_LIBOBOE = _oboe.OBOE_MODULE_LIBOBOE
OBOE_MODULE_SETTINGS = _oboe.OBOE_MODULE_SETTINGS
OBOE_MODULE_REPORTER_FILE = _oboe.OBOE_MODULE_REPORTER_FILE
OBOE_MODULE_REPORTER_UDP = _oboe.OBOE_MODULE_REPORTER_UDP
OBOE_MODULE_REPORTER_SSL = _oboe.OBOE_MODULE_REPORTER_SSL
OBOE_MODULE_APACHE = _oboe.OBOE_MODULE_APACHE
OBOE_MODULE_NGINX = _oboe.OBOE_MODULE_NGINX
OBOE_MODULE_PHP = _oboe.OBOE_MODULE_PHP
OBOE_MODULE_DOTNET = _oboe.OBOE_MODULE_DOTNET
class Metadata(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Metadata, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Metadata, name)
    __repr__ = _swig_repr

    def __init__(self, md):
        this = _oboe.new_Metadata(md)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_Metadata
    __del__ = lambda self: None
    if _newclass:
        fromString = staticmethod(_oboe.Metadata_fromString)
    else:
        fromString = _oboe.Metadata_fromString

    def createEvent(self):
        return _oboe.Metadata_createEvent(self)
    if _newclass:
        makeRandom = staticmethod(_oboe.Metadata_makeRandom)
    else:
        makeRandom = _oboe.Metadata_makeRandom

    def copy(self):
        return _oboe.Metadata_copy(self)

    def isValid(self):
        return _oboe.Metadata_isValid(self)

    def isSampled(self):
        return _oboe.Metadata_isSampled(self)

    def toString(self):
        return _oboe.Metadata_toString(self)
Metadata_swigregister = _oboe.Metadata_swigregister
Metadata_swigregister(Metadata)

def Metadata_fromString(s):
    return _oboe.Metadata_fromString(s)
Metadata_fromString = _oboe.Metadata_fromString

def Metadata_makeRandom(sampled=True):
    return _oboe.Metadata_makeRandom(sampled)
Metadata_makeRandom = _oboe.Metadata_makeRandom

class Context(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Context, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Context, name)
    __repr__ = _swig_repr
    if _newclass:
        setTracingMode = staticmethod(_oboe.Context_setTracingMode)
    else:
        setTracingMode = _oboe.Context_setTracingMode
    if _newclass:
        setDefaultSampleRate = staticmethod(_oboe.Context_setDefaultSampleRate)
    else:
        setDefaultSampleRate = _oboe.Context_setDefaultSampleRate
    if _newclass:
        sampleRequest = staticmethod(_oboe.Context_sampleRequest)
    else:
        sampleRequest = _oboe.Context_sampleRequest
    if _newclass:
        get = staticmethod(_oboe.Context_get)
    else:
        get = _oboe.Context_get
    if _newclass:
        toString = staticmethod(_oboe.Context_toString)
    else:
        toString = _oboe.Context_toString
    if _newclass:
        set = staticmethod(_oboe.Context_set)
    else:
        set = _oboe.Context_set
    if _newclass:
        fromString = staticmethod(_oboe.Context_fromString)
    else:
        fromString = _oboe.Context_fromString
    if _newclass:
        copy = staticmethod(_oboe.Context_copy)
    else:
        copy = _oboe.Context_copy
    if _newclass:
        setSampledFlag = staticmethod(_oboe.Context_setSampledFlag)
    else:
        setSampledFlag = _oboe.Context_setSampledFlag
    if _newclass:
        clear = staticmethod(_oboe.Context_clear)
    else:
        clear = _oboe.Context_clear
    if _newclass:
        isValid = staticmethod(_oboe.Context_isValid)
    else:
        isValid = _oboe.Context_isValid
    if _newclass:
        isSampled = staticmethod(_oboe.Context_isSampled)
    else:
        isSampled = _oboe.Context_isSampled
    if _newclass:
        init = staticmethod(_oboe.Context_init)
    else:
        init = _oboe.Context_init
    if _newclass:
        validateTransformServiceName = staticmethod(_oboe.Context_validateTransformServiceName)
    else:
        validateTransformServiceName = _oboe.Context_validateTransformServiceName
    if _newclass:
        shutdown = staticmethod(_oboe.Context_shutdown)
    else:
        shutdown = _oboe.Context_shutdown
    if _newclass:
        isReady = staticmethod(_oboe.Context_isReady)
    else:
        isReady = _oboe.Context_isReady
    if _newclass:
        createEvent = staticmethod(_oboe.Context_createEvent)
    else:
        createEvent = _oboe.Context_createEvent
    if _newclass:
        startTrace = staticmethod(_oboe.Context_startTrace)
    else:
        startTrace = _oboe.Context_startTrace

    def __init__(self):
        this = _oboe.new_Context()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_Context
    __del__ = lambda self: None
Context_swigregister = _oboe.Context_swigregister
Context_swigregister(Context)

def Context_setTracingMode(newMode):
    return _oboe.Context_setTracingMode(newMode)
Context_setTracingMode = _oboe.Context_setTracingMode

def Context_setDefaultSampleRate(newRate):
    return _oboe.Context_setDefaultSampleRate(newRate)
Context_setDefaultSampleRate = _oboe.Context_setDefaultSampleRate

def Context_sampleRequest(service_name, in_xtrace):
    return _oboe.Context_sampleRequest(service_name, in_xtrace)
Context_sampleRequest = _oboe.Context_sampleRequest

def Context_get():
    return _oboe.Context_get()
Context_get = _oboe.Context_get

def Context_toString():
    return _oboe.Context_toString()
Context_toString = _oboe.Context_toString

def Context_set(md):
    return _oboe.Context_set(md)
Context_set = _oboe.Context_set

def Context_fromString(s):
    return _oboe.Context_fromString(s)
Context_fromString = _oboe.Context_fromString

def Context_copy():
    return _oboe.Context_copy()
Context_copy = _oboe.Context_copy

def Context_setSampledFlag():
    return _oboe.Context_setSampledFlag()
Context_setSampledFlag = _oboe.Context_setSampledFlag

def Context_clear():
    return _oboe.Context_clear()
Context_clear = _oboe.Context_clear

def Context_isValid():
    return _oboe.Context_isValid()
Context_isValid = _oboe.Context_isValid

def Context_isSampled():
    return _oboe.Context_isSampled()
Context_isSampled = _oboe.Context_isSampled

def Context_init(*args):
    return _oboe.Context_init(*args)
Context_init = _oboe.Context_init

def Context_validateTransformServiceName(service_name):
    return _oboe.Context_validateTransformServiceName(service_name)
Context_validateTransformServiceName = _oboe.Context_validateTransformServiceName

def Context_shutdown():
    return _oboe.Context_shutdown()
Context_shutdown = _oboe.Context_shutdown

def Context_isReady(timeout):
    return _oboe.Context_isReady(timeout)
Context_isReady = _oboe.Context_isReady

def Context_createEvent():
    return _oboe.Context_createEvent()
Context_createEvent = _oboe.Context_createEvent

def Context_startTrace():
    return _oboe.Context_startTrace()
Context_startTrace = _oboe.Context_startTrace

class Event(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Event, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Event, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_destroy__ = _oboe.delete_Event
    __del__ = lambda self: None

    def addInfo(self, *args):
        return _oboe.Event_addInfo(self, *args)

    def addEdge(self, md):
        return _oboe.Event_addEdge(self, md)

    def addEdgeStr(self, val):
        return _oboe.Event_addEdgeStr(self, val)

    def getMetadata(self):
        return _oboe.Event_getMetadata(self)

    def metadataString(self):
        return _oboe.Event_metadataString(self)

    def send(self):
        return _oboe.Event_send(self)
    if _newclass:
        startTrace = staticmethod(_oboe.Event_startTrace)
    else:
        startTrace = _oboe.Event_startTrace
Event_swigregister = _oboe.Event_swigregister
Event_swigregister(Event)

def Event_startTrace(md):
    return _oboe.Event_startTrace(md)
Event_startTrace = _oboe.Event_startTrace

class Span(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Span, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Span, name)
    __repr__ = _swig_repr
    if _newclass:
        createSpan = staticmethod(_oboe.Span_createSpan)
    else:
        createSpan = _oboe.Span_createSpan
    if _newclass:
        createHttpSpan = staticmethod(_oboe.Span_createHttpSpan)
    else:
        createHttpSpan = _oboe.Span_createHttpSpan

    def __init__(self):
        this = _oboe.new_Span()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_Span
    __del__ = lambda self: None
Span_swigregister = _oboe.Span_swigregister
Span_swigregister(Span)

def Span_createSpan(transaction, domain, duration, service_name=None):
    return _oboe.Span_createSpan(transaction, domain, duration, service_name)
Span_createSpan = _oboe.Span_createSpan

def Span_createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name=None):
    return _oboe.Span_createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name)
Span_createHttpSpan = _oboe.Span_createHttpSpan

class MetricTags(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, MetricTags, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, MetricTags, name)
    __repr__ = _swig_repr

    def __init__(self, count):
        this = _oboe.new_MetricTags(count)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_MetricTags
    __del__ = lambda self: None

    def add(self, index, k, v):
        return _oboe.MetricTags_add(self, index, k, v)
MetricTags_swigregister = _oboe.MetricTags_swigregister
MetricTags_swigregister(MetricTags)

class CustomMetrics(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, CustomMetrics, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, CustomMetrics, name)
    __repr__ = _swig_repr
    if _newclass:
        summary = staticmethod(_oboe.CustomMetrics_summary)
    else:
        summary = _oboe.CustomMetrics_summary
    if _newclass:
        increment = staticmethod(_oboe.CustomMetrics_increment)
    else:
        increment = _oboe.CustomMetrics_increment

    def __init__(self):
        this = _oboe.new_CustomMetrics()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_CustomMetrics
    __del__ = lambda self: None
CustomMetrics_swigregister = _oboe.CustomMetrics_swigregister
CustomMetrics_swigregister(CustomMetrics)

def CustomMetrics_summary(name, value, count, host_tag, service_name, tags, tags_count):
    return _oboe.CustomMetrics_summary(name, value, count, host_tag, service_name, tags, tags_count)
CustomMetrics_summary = _oboe.CustomMetrics_summary

def CustomMetrics_increment(name, count, host_tag, service_name, tags, tags_count):
    return _oboe.CustomMetrics_increment(name, count, host_tag, service_name, tags, tags_count)
CustomMetrics_increment = _oboe.CustomMetrics_increment

class Reporter(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Reporter, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Reporter, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _oboe.new_Reporter(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_Reporter
    __del__ = lambda self: None

    def sendReport(self, *args):
        return _oboe.Reporter_sendReport(self, *args)

    def sendStatus(self, *args):
        return _oboe.Reporter_sendStatus(self, *args)
Reporter_swigregister = _oboe.Reporter_swigregister
Reporter_swigregister(Reporter)

class DebugLogger(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DebugLogger, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DebugLogger, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _oboe.delete_DebugLogger
    __del__ = lambda self: None

    def log(self, module, level, source_name, source_lineno, msg):
        return _oboe.DebugLogger_log(self, module, level, source_name, source_lineno, msg)
DebugLogger_swigregister = _oboe.DebugLogger_swigregister
DebugLogger_swigregister(DebugLogger)


def oboe_debug_log_handler(context, module, level, source_name, source_lineno, msg):
    return _oboe.oboe_debug_log_handler(context, module, level, source_name, source_lineno, msg)
oboe_debug_log_handler = _oboe.oboe_debug_log_handler
class DebugLog(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DebugLog, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DebugLog, name)
    __repr__ = _swig_repr
    if _newclass:
        getLevelName = staticmethod(_oboe.DebugLog_getLevelName)
    else:
        getLevelName = _oboe.DebugLog_getLevelName
    if _newclass:
        getModuleName = staticmethod(_oboe.DebugLog_getModuleName)
    else:
        getModuleName = _oboe.DebugLog_getModuleName
    if _newclass:
        getLevel = staticmethod(_oboe.DebugLog_getLevel)
    else:
        getLevel = _oboe.DebugLog_getLevel
    if _newclass:
        setLevel = staticmethod(_oboe.DebugLog_setLevel)
    else:
        setLevel = _oboe.DebugLog_setLevel
    if _newclass:
        setOutputStream = staticmethod(_oboe.DebugLog_setOutputStream)
    else:
        setOutputStream = _oboe.DebugLog_setOutputStream
    if _newclass:
        setOutputFile = staticmethod(_oboe.DebugLog_setOutputFile)
    else:
        setOutputFile = _oboe.DebugLog_setOutputFile
    if _newclass:
        addDebugLogger = staticmethod(_oboe.DebugLog_addDebugLogger)
    else:
        addDebugLogger = _oboe.DebugLog_addDebugLogger
    if _newclass:
        removeDebugLogger = staticmethod(_oboe.DebugLog_removeDebugLogger)
    else:
        removeDebugLogger = _oboe.DebugLog_removeDebugLogger
    if _newclass:
        logMessage = staticmethod(_oboe.DebugLog_logMessage)
    else:
        logMessage = _oboe.DebugLog_logMessage

    def __init__(self):
        this = _oboe.new_DebugLog()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_DebugLog
    __del__ = lambda self: None
DebugLog_swigregister = _oboe.DebugLog_swigregister
DebugLog_swigregister(DebugLog)

def DebugLog_getLevelName(level):
    return _oboe.DebugLog_getLevelName(level)
DebugLog_getLevelName = _oboe.DebugLog_getLevelName

def DebugLog_getModuleName(module):
    return _oboe.DebugLog_getModuleName(module)
DebugLog_getModuleName = _oboe.DebugLog_getModuleName

def DebugLog_getLevel(module):
    return _oboe.DebugLog_getLevel(module)
DebugLog_getLevel = _oboe.DebugLog_getLevel

def DebugLog_setLevel(module, newLevel):
    return _oboe.DebugLog_setLevel(module, newLevel)
DebugLog_setLevel = _oboe.DebugLog_setLevel

def DebugLog_setOutputStream(newStream):
    return _oboe.DebugLog_setOutputStream(newStream)
DebugLog_setOutputStream = _oboe.DebugLog_setOutputStream

def DebugLog_setOutputFile(pathname):
    return _oboe.DebugLog_setOutputFile(pathname)
DebugLog_setOutputFile = _oboe.DebugLog_setOutputFile

def DebugLog_addDebugLogger(newLogger, logLevel):
    return _oboe.DebugLog_addDebugLogger(newLogger, logLevel)
DebugLog_addDebugLogger = _oboe.DebugLog_addDebugLogger

def DebugLog_removeDebugLogger(oldLogger):
    return _oboe.DebugLog_removeDebugLogger(oldLogger)
DebugLog_removeDebugLogger = _oboe.DebugLog_removeDebugLogger

def DebugLog_logMessage(module, level, source_name, source_lineno, msg):
    return _oboe.DebugLog_logMessage(module, level, source_name, source_lineno, msg)
DebugLog_logMessage = _oboe.DebugLog_logMessage

class Config(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Config, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Config, name)
    __repr__ = _swig_repr
    if _newclass:
        checkVersion = staticmethod(_oboe.Config_checkVersion)
    else:
        checkVersion = _oboe.Config_checkVersion
    if _newclass:
        getVersion = staticmethod(_oboe.Config_getVersion)
    else:
        getVersion = _oboe.Config_getVersion
    if _newclass:
        getRevision = staticmethod(_oboe.Config_getRevision)
    else:
        getRevision = _oboe.Config_getRevision

    def __init__(self):
        this = _oboe.new_Config()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _oboe.delete_Config
    __del__ = lambda self: None
Config_swigregister = _oboe.Config_swigregister
Config_swigregister(Config)

def Config_checkVersion(version, revision):
    return _oboe.Config_checkVersion(version, revision)
Config_checkVersion = _oboe.Config_checkVersion

def Config_getVersion():
    return _oboe.Config_getVersion()
Config_getVersion = _oboe.Config_getVersion

def Config_getRevision():
    return _oboe.Config_getRevision()
Config_getRevision = _oboe.Config_getRevision

# This file is compatible with both classic and new-style classes.





#_*_ coding: utf-8 _*_
""" AppOptics APM instrumentation API for Python.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.

appoptics_apm_test defines test mock classes for:
a) running unit test
"""
import pprint
import socket
import appoptics_apm
import string
import random

#objects pickle unpickle with HAC(hash authentication)
try:
    import cPickle as pickle
except ImportError:
    import pickle

# global set of registered event reporting listeners
listeners = []


class Metadata(object):
    def __init__(self, flag=None):
        if flag:
            int_randomer = random.Random()
            id_list = [
                string.hexdigits[int_randomer.randint(0, len(string.hexdigits) -1)]
                for _ in range(56)]
            self.id = "2B" + ''.join(id_list).upper() + '01'
        else:
            self.id = ''

    @staticmethod
    def fromString(strv):
        md = Metadata()
        md.id = strv
        return md

    def createEvent(self):
        return Event()

    @staticmethod
    def makeRandom(flag=True):
        return Metadata(flag)

    def copy(self):
        return self

    def isValid(self):
        return True

    def isSampled(self):
        return True

    def toString(self):
        return self.id


class Context(object):
    md = None

    @staticmethod
    def init(_, __, ___):
        pass

    @staticmethod
    def setTracingMode(_):
        return False

    @staticmethod
    def setDefaultSampleRate(_):
        return False

    @staticmethod
    def sampleRequest(_, __):
        return True

    @classmethod
    def get(cls):
        return cls.md

    @classmethod
    def set(cls, md):
        cls.md = md

    @staticmethod
    def fromString(_):
        return Context()

    @staticmethod
    def copy():
        return Context()

    @classmethod
    def clear(cls):
        cls.md = None

    @classmethod
    def isValid(cls):
        return cls.md != None

    @classmethod
    def isSampled(cls):
        return cls.md != None

    @staticmethod
    def toString():
        return str(Context.md.id) if Context.md else ''

    @staticmethod
    def createEvent():
        return Event()

    @staticmethod
    def startTrace(_=None):
        return Event()


class Event(object):
    def __init__(self, _=None, __=None):
        ''' initialize event. Props for checking properties,
            evt for compatible with oboe SwigEvent. Otherwise
            complaints come from _reporter().sendReport(event.evt)
            etc.
        '''
        self.props = {}
        self.evt = self

    def addInfo(self, name, value):
        self.props[name] = value

    def addEdge(self, _):
        pass

    def addEdgeStr(self, _):
        pass

    def getMetadata(self):
        return Metadata()

    def metadataString(self):
        return ''

    def is_valid(self):
        return True

    @staticmethod
    def startTrace(_=None):
        return Event()

    def __repr__(self):
        return '<Event('+str(self.props)+')>'

# Agent reporter_instance is initialized the first time it is used
# By design, it uses polymorphism to initialize corresponding reporter
# instance for oboe. For python, it should use factory or like this
# use proxy to address that

class Reporter(object):
    """ Mock the Reporter; no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, mode, config_str):
        ''' initialize a udp reporter and helper functions for
            network data processing if it's configured appropriately
        :param  mode:str, reporter mode, should be udp
        :param config_str:str, port, and cid. Borrowed from
           oboe, port = server_host:listen_port, cid = dummy,
           i.e 'port=localhost:8000,cid=you-key'
        :param timeout:int timeout for reading data
        '''
        if mode == "udp":
            try:
                host = config_str.split(",")[0].split("=")[1].split(":")[0]
                port = int(config_str.split(",")[0].split("=")[1].split(":")[1])
                self.proxy = UdpReporter(mode, (host, port))
            except:
                self.proxy = None
        else:
            self.proxy = None

    def sendReport(self, event, __=None):
        if self.proxy:
            self.proxy.sendReport(event)
            return
        for listener in listeners:
            listener.send(event)

    def sendStatus(self, event, __=None):
        if self.proxy:
            self.proxy.sendStatus(event)
            return
        for listener in listeners:
            listener.send(event)

    def sendSpan(self, event, __=None):
        if self.proxy:
            self.proxy.sendSpan(event)
            return
        for listener in listeners:
            listener.send(event)

class UdpReporter(object):
    """ Reporter for no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, mode="udp", server_tuple=('localhost', 7835), timeout=30):
        ''' initialize a upd reporter and helper functions for
        network data processing
        :param  mode:str, reporter mode, should be udp
        :param server_tuple:tuple, server address and port, default 7835 is defined
               in net_utils.py
        :param timeout:int timeout for reading data
        '''
        self.listener = (server_tuple[0], server_tuple[1])
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.settimeout(timeout)

    def _pickle_message(self, obj):
        '''encode an object for network transfer'''
        return pickle.dumps(obj)

    def sendReport(self, event, __=None):
        self.socket.sendto(self._pickle_message(event), self.listener)

    def sendStatus(self, event, __=None):
        self.socket.sendto(self._pickle_message(event), self.listener)

    def sendSpan(self, event, __=None):
        self.socket.sendto(self._pickle_message(event), self.listener)

class SslReporter(object):
    """ Mock UDP Reported; no-op for unsupported platforms, or unit test harness
        if in APPOPTICS_TEST mode. """
    def __init__(self, _, __=None):
        pass

    def sendReport(self, event, __=None):
        for listener in listeners:
            listener.send(event)

    def sendStatus(self, event, __=None):
        for listener in listeners:
            listener.send(event)

    def sendSpan(self, event, __=None):
        if self.proxy:
            self.proxy.sendSpan(event)
            return
        for listener in listeners:
            listener.send(event)

class Span(object):
    @staticmethod
    def createHttpSpan(*args):
        #util._reporter() guarantees reporter instance
        reporter = appoptics_apm.util._reporter()
        if reporter:
            reporter.sendSpan({'SPAN_REPORT': args})

    @staticmethod
    def createSpan(*args):
        #util._reporter() guarantees reporter instance
        reporter = appoptics_apm.util._reporter()
        if reporter:
            reporter.sendSpan({'SPAN_REPORT': args})

class MetricTags(object):
    def add(*args, **kwargs):
        pass


class CustomMetrics(object):
    @staticmethod
    def summary(*args, **kwargs):
        pass

    @staticmethod
    def increment(*args, **kwargs):
        pass


class oboe_metric_tag_t(object):
    def __init__(self, k, v):
        self.key = k
        self.value = v


class AppOpticsApmListener(object):
    """ Simple test harness for intercepting event reports. """
    def __init__(self):
        self.events = []
        self.spans = []
        self.listeners = listeners
        listeners.append(self)

    def send(self, event):
        if isinstance(event, Event):
            self.events.append(event)
        else:
            self.spans.append(event)

    def get_events(self, *filters):
        """ Returns all events matching the filters passed """
        events = self.events
        for _filter in filters:
            events = [ev for ev in events if _filter(ev)]
        return events

    def get_spans(self, *filters):
        """ Returns all spans matching the filters passed """
        spans = self.spans
        for _filter in filters:
            spans = [sp for sp in spans if _filter(sp)]
        return spans

    def str_events(self, *filters):
        return pprint.pformat(self.get_events(*filters))

    def str_spans(self, *filters):
        return pprint.pformat(self.get_spans(*filters))

    def pop_events(self, *filters):
        """ Returns all events matching the filters passed,
        and also removes those events from the Trace so that
        they will not be returned by future calls to
        pop_events or events. """
        matched = self.get_events(*filters)
        for match in matched:
            self.events.remove(match)
        return matched

    def pop_spans(self, *filters):
        """ Returns all spans matching the filters passed,
        and also removes those spans from the Trace so that
        they will not be returned by future calls to
        pop_spans or spans. """
        matched = self.get_spans(*filters)
        for match in matched:
            self.spans.remove(match)
        return matched

    def __del__(self):
        self.listeners.remove(self)


class DebugLog(object):
    @staticmethod
    def getLevelName(*args, **kwargs):
        pass

    @staticmethod
    def getModuleName(*args, **kwargs):
        pass

    @staticmethod
    def getLevel(*args, **kwargs):
        pass

    @staticmethod
    def setLevel(*args, **kwargs):
        pass

    @staticmethod
    def setOutputStream(*args, **kwargs):
        pass

    @staticmethod
    def setOutputFile(*args, **kwargs):
        pass

    @staticmethod
    def addDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def removeDebugLogger(*args, **kwargs):
        pass

    @staticmethod
    def logMessage(*args, **kwargs):
        pass



""" useful code for instrumenting asynchronous Python programs
# (c) 2016 SolarWinds, LLC.
"""
from appoptics_apm import util


class AppOpticsApmContextManager(object):
    """ A context manager (for the "with" statement) that sets and
        clears the context when entered, storing the metadata in an
        object passed by the constructor.  E.g.:

        with AppOpticsApmContextManager(self.request):
            do_something()

        Here, any code called from do_something() will have its appoptics_apm
        context set from (and stored to, after finishing) the
        self.request object.
    """

    def __init__(self, ctxobj=None):
        """ stores appoptics_apm metadata as attribute of object 'ctxobj' """
        self.obj = ctxobj

    def __enter__(self):
        ctx = getattr(self.obj, '_appoptics_apm_ctx', None)
        if ctx and ctx.is_valid():
            ctx.set_as_default()
        elif util.Context.get_default().is_valid():
            util.Context.clear_default()

    def __exit__(self, type, value, tb):
        default_ctx = util.Context.get_default()
        if default_ctx.is_valid():
            ctx = default_ctx.copy()
            setattr(self.obj, '_appoptics_apm_ctx', ctx)
            util.Context.clear_default()



""" AppOptics APM instrumentation for Django ORM.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""

from appoptics_apm import util
import re

appoptics_apm_logger = util.logger


def wrap_execute(func, f_args, f_kwargs, res):
    obj, sql = f_args[:2]
    kwargs = {}
    log_sql_args = not util.config.get('enable_sanitize_sql', True) and len(f_args) > 2
    if log_sql_args:
        kwargs['QueryArgs'] = str(f_args[2]).encode('utf-8')

    kwargs['Query'] = sql.encode('utf-8')
    if 'NAME' in obj.db.settings_dict:
        kwargs['Database'] = obj.db.settings_dict['NAME']
    if 'HOST' in obj.db.settings_dict:
        kwargs['RemoteHost'] = obj.db.settings_dict['HOST']
    if 'ENGINE' in obj.db.settings_dict:
        if re.search('post', obj.db.settings_dict['ENGINE']):
            kwargs['Flavor'] = 'postgresql'
        elif re.search('sqlite', obj.db.settings_dict['ENGINE']):
            kwargs['Flavor'] = 'sqlite'
        elif re.search('mysql', obj.db.settings_dict['ENGINE']):
            kwargs['Flavor'] = 'mysql'
        elif re.search('oracle', obj.db.settings_dict['ENGINE']):
            kwargs['Flavor'] = 'oracle'
    return kwargs


class CursorAppOpticsApmWrapper(object):
    ###########################################################################
    # Django cursors can be wrapped arbitrarily deeply with the following API.
    # Each class contains a references to the DB object, and the next level
    # cursor. Control passes to the cursor in execute and executemany, wrapped
    # with whatever behavior the wrapper provides.
    ###########################################################################

    def __init__(self, cursor, db):
        self.cursor = cursor
        self.db = db

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return getattr(self.cursor, attr)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __iter__(self):
        return iter(self.cursor)

    def execute(self, sql, params=None):
        if params is None:
            return self.cursor.execute(sql)
        else:
            return self.cursor.execute(sql, params)

    def executemany(self, sql, param_list):
        return self.cursor.executemany(sql, param_list)


def wrap(module):
    try:
        cursor_method = module.BaseDatabaseWrapper.cursor
        if getattr(cursor_method, '_appoptics_apm_wrapped', False):
            return

        appoptics_apm_wrapper = util.log_method(
            'djangoORM', callback=wrap_execute,
            store_backtrace=util._collect_backtraces('django_orm')
        )
        setattr(CursorAppOpticsApmWrapper, 'execute', appoptics_apm_wrapper(CursorAppOpticsApmWrapper.execute))
        setattr(CursorAppOpticsApmWrapper, 'executemany', appoptics_apm_wrapper(CursorAppOpticsApmWrapper.executemany))

        def cursor_wrap(self):
            try:
                return CursorAppOpticsApmWrapper(cursor_method(self), self)
            except Exception as e:
                appoptics_apm_logger.error("[AppOptics] Error in cursor_wrap: %s" % e)
                # need to re-raise application exceptions thrown here
                raise

        cursor_wrap._appoptics_apm_wrapped = True

        setattr(module.BaseDatabaseWrapper, 'cursor', cursor_wrap)
    except Exception as e:
        appoptics_apm_logger.error("[AppOptics] Error in module_wrap: %s" % e)
